# coding: utf-8
from __future__ import unicode_literals

from django.db.models.query import QuerySet
from django.contrib.auth.models import User, UserManager
from django.test.utils import override_settings

from .base import BaseTest
from granular_access.queryset import AccessQuerySet, AccessQuerySetMixin
from granular_access.manager import AccessManager, AccessManagerMixin

class AccessQuerySetTest(BaseTest):
    def setUp(self):
        self.batman = User.objects.create_user(username='batman',
                                               email='batman@gotham.us')

    def test_queryset(self):
        queryset = AccessQuerySet(model=User)

        filtered = queryset.available(to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, []
        )

    @override_settings(CONSIDER_SUPERUSER=True)
    def test_queryset_superuser_enabled(self):
        self.batman.is_superuser = True
        self.batman.save()

        queryset = AccessQuerySet(model=User)

        filtered = queryset.available(to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, [self.batman]
        )

class AccessQuerySetMixinTest(BaseTest):
    class MyCustomQuerySet(QuerySet, AccessQuerySetMixin):
        pass

    def setUp(self):
        self.batman = User.objects.create_user(username='batman',
                                               email='batman@gotham.us')

    def test_queryset_mixin(self):
        queryset = self.MyCustomQuerySet(model=User)

        filtered = queryset.available(to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, []
        )

    @override_settings(CONSIDER_SUPERUSER=True)
    def test_queryset_mixin_superuser_enabled(self):
        self.batman.is_superuser = True
        self.batman.save()

        queryset = self.MyCustomQuerySet(model=User)

        filtered = queryset.available(to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, [self.batman]
        )

class AccessManangerTest(BaseTest):
    class UserAccessProxy(User):
        objects = AccessManager()
        class Meta:
            proxy = True

    def setUp(self):
        self.batman = User.objects.create_user(username='batman',
                                               email='batman@gotham.us')

    def test_manager(self):
        filtered = self.UserAccessProxy.objects.available(
            to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, []
        )

    @override_settings(CONSIDER_SUPERUSER=True)
    def test_manager_superuser_enabled(self):
        self.batman.is_superuser = True
        self.batman.save()

        filtered = self.UserAccessProxy.objects.available(
            to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, [self.batman]
        )

class AccessManangerMixinTest(BaseTest):

    class UserAccessProxy(User):
        class MyCustomAccessManager(UserManager, AccessManagerMixin):
            pass

        objects = MyCustomAccessManager()
        class Meta:
            proxy = True

    def setUp(self):
        self.batman = User.objects.create_user(username='batman',
                                               email='batman@gotham.us')

    def test_manager(self):
        filtered = self.UserAccessProxy.objects.available(
            to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, []
        )

    @override_settings(CONSIDER_SUPERUSER=True)
    def test_manager_superuser_enabled(self):
        self.batman.is_superuser = True
        self.batman.save()

        filtered = self.UserAccessProxy.objects.available(
            to=self.batman, action='edit')

        self.assertQuerysetEqualsList(
            filtered, [self.batman]
        )



# coding: utf-8
from __future__ import unicode_literals

from django.db import models

from .settings import USER_MODEL, GROUP_MODEL

class ACL(models.Model):
    """Access control list"""
    user = models.ForeignKey(
        to=USER_MODEL,
        verbose_name='user',
        null=True, blank=True,
    )
    group = models.ForeignKey(
        to=GROUP_MODEL,
        verbose_name='Group',
        null=True, blank=True
    )
    lookup = models.ForeignKey(
        to='stored_filters.Filter',
        verbose_name='lookup',
    )
    action = models.SlugField(
        verbose_name='action',
        blank=True,
    )

    def __unicode__(self):
        if self.user is None and self.group is None:
            who = 'everybody'
        elif self.user is not None and self.group is None:
            who = unicode(self.user)
        elif self.user is None and self.group is not None:
            who = unicode(self.group)
        else:
            who = '%s and %s' % (self.user, self.group)

        if self.action:
            what = unicode(self.action)
        else:
            what = 'do anything with'

        objects = unicode(self.lookup)

        return '%s can %s %s' % (who, what, objects)




# coding: utf-8
from __future__ import unicode_literals

from django.conf import settings

USER_MODEL = getattr(settings, 'GRANULAR_ACCESS_USER_MODEL', 'auth.User')
GROUP_MODEL = getattr(settings, 'GRANULAR_ACCESS_GROUP_MODEL', 'auth.Group')

USER_GROUP_RELATED_NAME = getattr(settings,
    'GRANULAR_ACCESS_USER_GROUP_RELATED_NAME', 'groups')

GET_USER_GROUPS_FUNCTION = getattr(settings,
    'GRANULAR_ACCESS_GET_USER_GROUPS_FUNCTION', None)

CONSIDER_SUPERUSER = getattr(settings,
    'GRANULAR_ACCESS_CONSIDER_SUPERUSER', True)




# coding: utf-8
from __future__ import unicode_literals
from django.test import TestCase

class BaseTest(TestCase):
    def assertQuerysetEqualsList(self, queryset, list_):
        self.assertEqual(set(queryset.values_list('id', flat=True)),
                         set(obj.id for obj in list_))



# coding: utf-8
from __future__ import unicode_literals
from django.core import exceptions
from django.utils.importlib import import_module

def get_function_by_path(path):
    try:
        module_name, function_name = path.rsplit('.', 1)
    except ValueError:
        raise exceptions.ImproperlyConfigured(
            '%s isn\'t a function module_name' % path)
    try:
        module = import_module(module_name)
    except ImportError, e:
        raise exceptions.ImproperlyConfigured(
            'Error importing module_name %s: "%s"' % (module_name, e))
    try:
        function = getattr(module, function_name)
    except AttributeError:
        raise exceptions.ImproperlyConfigured(
            'Module "%s" does not define a "%s" function' % (
                module_name, function_name))
    return function



# coding: utf-8

# TODO: reverse get

from __future__ import unicode_literals
from django.contrib.contenttypes.models import ContentType
from django.forms.models import model_to_dict

from stored_filters import create_filter, filters_as_Q

from .models import ACL
from .acl import match_acl
from stored_filters.query import QueryBuilder


def filter_available(to, action, queryset):
    """Filter objects, available to user for specified action.
    to -- user instance
    action -- action name
    queryset -- base queryset for filtering

    """
    app_label = queryset.model._meta.app_label
    model_name = queryset.model._meta.module_name
    query = get_filter_query(user=to, action=action,
                             app_label=app_label, model_name=model_name)
    if query is None:
        return queryset.none()
    return queryset.filter(query)

def has_perm(user, action, instance):
    options = instance._meta
    model = ContentType.objects.get_by_natural_key(
        app_label=options.app_label, model=options.module_name).model_class()
    queryset = model.objects.all()
    available = filter_available(to=user, action=action, queryset=queryset)
    return available.filter(pk=instance.pk).exists()

def get_filter_query(user, action, app_label, model_name):
    """Get query as Q object for filtering objects.
    to -- user instance
    action -- action name
    app_label -- model's app_label
    model_name -- model's name

    @return Q-object or None
    """
    user_acls = match_acl(user, action, app_label, model_name)
    lookups = [acl.lookup for acl in user_acls]
    user_as_dict = model_to_dict(user)
    return filters_as_Q(filters=lookups, context=user_as_dict)

def create_permission(user=None, group=None, action='',
                      conditions=None, exclusions=None,
                      model_class=None, app_label=None, model_name=None):

    lookup = create_filter(conditions, exclusions,
        model_class, app_label, model_name)

    ACL.objects.create(user=user, group=group, action=action, lookup=lookup)

def get_permissions_owner(model_class=None, app_label=None, model_name=None,
                          action=None):
    pass




# coding: utf-8
from __future__ import unicode_literals

from .access import get_filter_query, filter_available, create_permission
from .manager import AccessManager, AccessManagerMixin
from .queryset import AccessQuerySet, AccessQuerySetMixin



# coding: utf-8
from __future__ import unicode_literals
from django.db.models import Q

from .models import ACL
from .settings import  GET_USER_GROUPS_FUNCTION as GET_GROUPS
from .settings import  USER_GROUP_RELATED_NAME as RELATED_NAME
from .utils import get_function_by_path

def match_acl(user, action, app_label, model_name):
    user_groups = get_user_groups(user=user)

    user_query = (
        Q(user=user) |
        Q(group__in=user_groups) |
        Q(user__isnull=True) & Q(group__isnull=True)
    )
    model_query = (
        Q(lookup__content_type__model=model_name) &
        Q(lookup__content_type__app_label=app_label)
    )
    action_query = Q(action=action) | Q(action='')

    return ACL.objects.select_related('lookup').filter(
        user_query, model_query, action_query)

def get_user_groups(user):
    if GET_GROUPS is None:
        groups_related_manager = getattr(user, RELATED_NAME)
        return groups_related_manager.all()
    else:
        group_getter = get_function_by_path(GET_GROUPS)
        return group_getter(user)




# coding: utf-8
from __future__ import unicode_literals

from django.contrib.auth.models import User, Group

from stored_filters import create_filter
from granular_access.models import ACL
from granular_access.tests.base import BaseTest
from granular_access.acl import match_acl

class MatchACLTest(BaseTest):
    def setUp(self):
        self.batman = User.objects.create_user(username='batman')
        self.joker = User.objects.create_user(username='joker')
        self.superheroes = Group.objects.create(name='superheroes')

        self.lookup = create_filter(model_class=User)

    def test_user_personal_acl(self):
        acl1 = ACL.objects.create(user=self.batman, group=None,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_group_acl(self):
        self.superheroes.user_set.add(self.batman)
        acl1 = ACL.objects.create(user=None, group=self.superheroes,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_kill_and_rob_acl(self):
        acl1 = ACL.objects.create(user=None, group=None,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_user_or_group_acl(self):
        """weird looking, but should work"""
        acl1 = ACL.objects.create(user=self.batman, group=self.superheroes,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

        self.superheroes.user_set.add(self.batman)
        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_permissionless_user(self):
        ACL.objects.create(user=self.batman, group=None,
                           lookup=self.lookup, action='')

        matched = match_acl(user=self.joker, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [])

    def test_action_match(self):
        acl1 = ACL.objects.create(user=None, group=None,
                                  lookup=self.lookup, action='kill-rob')

        matched = match_acl(user=self.joker, action='kill-rob',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_action_mismatch(self):
        ACL.objects.create(user=None, group=None,
                           lookup=self.lookup, action='kill-rob')

        matched = match_acl(user=self.joker, action='save',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [])

    def test_action_wildcard_match(self):
        acl1 = ACL.objects.create(user=None, group=None,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.joker, action='kill-rob',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_model_match(self):
        acl1 = ACL.objects.create(user=None, group=None,
                                  lookup=self.lookup, action='')

        matched = match_acl(user=self.joker, action='kill-rob',
                            model_name='user', app_label='auth')

        self.assertQuerysetEqualsList(matched, [acl1])

    def test_model_mismatch(self):
        ACL.objects.create(user=None, group=None,
                            lookup=self.lookup, action='')

        matched = match_acl(user=self.joker, action='kill-rob',
                            model_name='group', app_label='auth')

        self.assertQuerysetEqualsList(matched, [])

    def test_full_case(self):
        self.superheroes.user_set.add(self.batman)

        acl1 = ACL.objects.create(user=self.batman, group=None,
                                  lookup=self.lookup, action='save')
        acl2 = ACL.objects.create(user=None, group=self.superheroes,
                                  lookup=self.lookup, action='save')
        acl3 = ACL.objects.create(user=self.joker, group=None,
                                  lookup=self.lookup, action='kill-rob')

        matched = match_acl(user=self.batman, action='save',
                            model_name='user', app_label='auth')
        self.assertQuerysetEqualsList(matched, [acl1, acl2])

        matched = match_acl(user=self.batman, action='kill-rob',
                            model_name='user', app_label='auth')
        self.assertQuerysetEqualsList(matched, [])

        matched = match_acl(user=self.joker, action='kill-rob',
                            model_name='user', app_label='auth')
        self.assertQuerysetEqualsList(matched, [acl3])



# coding: utf-8
from __future__ import unicode_literals

from django.test import TestCase

from granular_access.utils import get_function_by_path

def get_groups(user):
    return [1, 2, 3]

class UtilsTest(TestCase):

    FUNCTION_PATH = 'granular_access.tests.utils.get_groups'

    def test_get_function_by_path(self):
        function = get_function_by_path(path=self.FUNCTION_PATH)
        self.assertEqual(
            function({}),
            [1, 2, 3]
        )



# coding: utf-8
from __future__ import unicode_literals

from django.db.models import Manager

from .queryset import AccessQuerySet, check_superuser
from .access import filter_available

class AccessManager(Manager):
    def get_query_set(self):
        return AccessQuerySet(model=self.model, using=self._db)

    def available(self, *args, **kwargs):
        return self.get_query_set().available(*args, **kwargs)

class AccessManagerMixin(object):
    def available(self, to, action):
        queryset = self.get_query_set()
        if check_superuser():
            return queryset
        return filter_available(to=to, action=action, queryset=queryset)



# coding: utf-8
from __future__ import unicode_literals
from django.db.models.query import QuerySet

from .access import filter_available
from .settings import CONSIDER_SUPERUSER

class AccessQuerySet(QuerySet):
    def available(self, to, action):
        if check_superuser(user=to):
            return self
        return filter_available(to=to, action=action, queryset=self)

class AccessQuerySetMixin(object):
    def available(self, to, action):
        if check_superuser(user=to):
            return self
        return filter_available(to=to, action=action, queryset=self)

def check_superuser(user):
    if CONSIDER_SUPERUSER:
        return user.is_superuser
    return False



import os
from distutils.core import setup


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
README_FILEPATH = os.path.join(CURRENT_DIR, 'README.rst')


setup(
    name='django-granular-access',
    version='0.1.1',
    packages=[
        'granular_access',
        'granular_access.tests',
        'granular_access.migrations'
    ],
    author='Kirill Sibirev',
    author_email='k.sibirev@gmail.com',
    url='https://github.com/l0kix2/django-granular-access',
    description='Flexible permission system for Django.',
    long_description=open(README_FILEPATH).read(),
    keywords=['django', 'permissions'],
    license='MIT',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
)



# coding: utf-8
from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType

from .base import BaseTest

from stored_filters import create_filter
from granular_access.models import ACL
from granular_access.access import get_filter_query, filter_available
from granular_access.access import create_permission, has_perm

class PublicApiTest(BaseTest):
    def setUp(self):
        self.batman = User.objects.create_user(username='batman',
                                               email='batman@gotham.us')
        self.joker = User.objects.create_user(username='joker',
                                              email='joker@gotham.us')
        self.god = User.objects.create_user(username='god',
                                            email='god@gmail.com')

        lookup = create_filter(model_class=User,
            conditions=[{'username__startswith': 'joker'}])
        ACL.objects.create(user=self.batman, group=None,
                           lookup=lookup, action='edit')

        blank_lookup = create_filter(model_class=User,
            conditions=None, exclusions=None)
        ACL.objects.create(user=self.joker, group=None,
                           lookup=blank_lookup, action='edit')

        god_lookup = create_filter(model_class=User,
            conditions={}, exclusions=None)
        ACL.objects.create(user=self.god, group=None,
                           lookup=god_lookup, action='edit')

    def test_get_filter_query(self):
        query = get_filter_query(user=self.batman, action='edit',
                                 app_label='auth', model_name='user')

        self.assertQuerysetEqualsList(
            User.objects.filter(query),
            [self.joker]
        )

    def test_get_filter_query_blank_lookups(self):


        query = get_filter_query(user=self.joker, action='edit',
                                 app_label='auth', model_name='user')

        self.assertIsNone(query)

    def test_filter_available(self):
        filtered = filter_available(to=self.batman, action='edit',
                                    queryset=User.objects.all())

        self.assertQuerysetEqualsList(
            filtered,
            [self.joker]
        )

    def test_has_perm(self):
        self.assertTrue(
            has_perm(user=self.batman, action='edit', instance=self.joker)
        )

    def test_has_no_perm(self):
        self.assertFalse(
            has_perm(user=self.batman, action='edit', instance=self.god)
        )

    def test_filter_available_blank_query(self):
        filtered = filter_available(to=self.joker, action='edit',
                                    queryset=User.objects.all())

        self.assertQuerysetEqualsList(filtered, [])

    def test_filter_available_super_query(self):
        filtered = filter_available(to=self.god, action='edit',
                                    queryset=User.objects.all())

        self.assertQuerysetEqualsList(
            filtered,
            [self.god, self.batman, self.joker]
        )

    def test_create_permission_using_model(self):
        conditions = [{'some_condition': 'value'}]
        action = 'some_action'

        create_permission(user=self.batman, action=action,
            conditions=conditions, model_class=User)

        user_content_type = ContentType.objects.get_for_model(User)
        self.assertTrue(
            ACL.objects.filter(action=action, user=self.batman,
                lookup__content_type=user_content_type).exists()
        )

    def test_create_permission_using_ct_natural_key(self):
        conditions = [{'some_condition': 'value'}]
        action = 'some_action'

        create_permission(user=self.batman, action=action,
            conditions=conditions, app_label='auth', model_name='user')

        user_content_type = ContentType.objects.get_for_model(User)
        self.assertTrue(
            ACL.objects.filter(action=action, user=self.batman,
                lookup__content_type=user_content_type).exists()
        )



# coding: utf-8
from __future__ import unicode_literals

from .acl import MatchACLTest
from .public_api import PublicApiTest
from .utils import UtilsTest
from .manager import AccessQuerySetTest, AccessQuerySetMixinTest
from .manager import AccessManangerTest, AccessManangerMixinTest

